

crafting_recipes <-
{
  fireball={
    type="spell"
    name="FireballSpell"
    tier=0
    parents=[]
    costs={ firebloom=1 }}

  meteor={
    type="spell"
    name="MeteorSpell"
    tier=1
    parents=["fireball"]
    costs={ shard=10 firebloom=2 vortex=1 }}

  lightning={
    type="spell"
    name="LightningSpell"
    tier=0
    parents=["fireball"]
    costs={ shard=5 vortex=5 }}

  thunderstorm={
    type="spell"
    name="ThunderstormSpell"
    tier=2
    parents=["lightning"]
    costs={ shard=20 star=5 seed=1 vortex=3 }}

  castle={
    type="spell"
    name="CastleSpell"
    tier=0
    parents=[]
    costs={ shard=1 vortex=1 seed=1 }}

  fireballTower={
    type="spell"
    name="GuardTowerSpell"
    tier=1
    parents=["castle", "fireball"]
    costs={ shard=10 firebloom=5 }}

  // lightningTower={
  //   type="spell"
  //   name="LightningTowerSpell"
  //   tier=2
  //   parents=["fireballTower", "lightning"]
  //   costs={ shard=1 }}

  waterGlobe={
    type="spell"
    name="WaterballSpell"
    tier=0
    parents=[]
    costs={ shard=2 vortex=10 }}

  waterSink={
    type="spell"
    name="WaterSinkSpell"
    tier=1
    parents=["waterGlobe"]
    costs={ vortex=10 seed=2 }}

  waterSource={
    type="spell"
    name="WaterSourceSpell"
    tier=1
    parents=["waterGlobe"]
    costs={ vortex=10 seed=5 }}

  healing={
    type="spell"
    name="HealingSpell"
    tier=0
    parents=[]
    costs={ vortex=5 seed=1 }}

  // shield={
  //   type="spell"
  //   name="ShieldSpell"
  //   tier=1
  //   parents=["healing"]
  //   costs={ shard=1 }}

  // invisibility={
  //   type="spell"
  //   name="InvisibilitySpell"
  //   tier=2
  //   parents=["shield"]
  //   costs={ shard=1 }}

  // claimMana={
  //   type="spell"
  //   name="ClaimManaSpell"
  //   tier=0
  //   parents=[]
  //   costs={ shard=1 }}

  // manaMagnet={
  //   type="spell"
  //   name="ManaMagnetSpell"
  //   tier=1
  //   parents=["claimMana"]
  //   costs={ shard=1 }}

  pit={
    type="spell"
    name="PitSpell"
    tier=0
    parents=[]
    costs={ shard=10 vortex=2 }}

  hill={
    type="spell"
    name="HillSpell"
    tier=1
    parents=["pit"]
    costs={ shard=10 firebloom=1 }}

  // tremor={
  //   type="spell"
  //   name="TremorSpell"
  //   tier=1
  //   parents=["pit"]
  //   costs={ shard=1 }}

  fracture={
    type="spell"
    name="FaultSpell"
    tier=2
    parents=["hill"]
    costs={ shard=10 star=1 firebloom=1 }}

  volcano={
    type="spell"
    name="VolcanoSpell"
    tier=3
    parents=["fracture"]
    costs={ shard=20 firebloom=20 star=5 }}
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


function was_recipe_crafted(name)
{
  if (!(name in crafting_recipes)) return false;

  local r=crafting_recipes[name]
  if (r.type=="spell")
  {
    return player_has_spell(r.name);
  }
  else
    return false;
}


function is_crafting_recipe_available(name)
{
  if (!(name in crafting_recipes)) return false;

  local r=crafting_recipes[name]
  if (r.type=="spell")
  {
    if (player_has_spell(r.name)) return false;
    foreach (p in r.parents) if (!was_recipe_crafted(p)) return false;
    return true;
  }
  else
    return false;
}


function get_next_crafting_recipes(num)
{
  local available=[]
  foreach (n,r in crafting_recipes)
    if (is_crafting_recipe_available(n)) available.append(r)

  available.sort(@(a, b) a.tier<=>b.tier)

  set_rnd_seed(universe.craftingSeed)

  local recipes=[], curTierNum=0
  while (recipes.len()<num && available.len()>0)
  {
    if (curTierNum<=0)
    {
      curTierNum=1
      while (curTierNum<available.len()
        && available[curTierNum].tier==available[0].tier)
          ++curTierNum;
    }

    local i=irnd()%curTierNum
    recipes.append(available[i])
    available.remove(i)
    --curTierNum
  }

  randomize_rnd_seed()

  return recipes;
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


function sim_crafting(rc, crafted, resources, result)
{
  local resClusters=[
    { shard=20 vortex=25 seed=1 }
    { firebloom=47 vortex=22 shard=78 seed=21 star=11 }
  ]

  // find craftable recipes
  local available=[]
  foreach (n,r in crafting_recipes)
  {
    if (n in crafted) continue;

    local ok=true
    foreach (p in r.parents) if (!(p in crafted)) { ok=false; break; }
    if (!ok) continue;

    foreach (rn,c in r.costs) if (resources[rn]<c) { ok=false; break; }
    if (!ok) continue;

    available.append(n)
  }

  if (!available.len())
  {
    // if no recipes, go to next cluster
    local cn=[]
    foreach (n,v in crafted) cn.append(n)
    cn.sort()

    // record variant
    local vn=""
    foreach (n in cn) vn+=" "+n

    if (!(vn in result) || result[vn]>rc)
      result[vn]<-rc

    if (rc>=resClusters.len()) return;

    local res=clone resources
    foreach (n,v in resClusters[rc]) res[n]+=v
    sim_crafting(rc+1, crafted, res, result)
  }
  else
  {
    // enumerate recipes
    foreach (rcn in available)
    {
      local r=crafting_recipes[rcn]
      local cr=clone crafted
      local res=clone resources
      cr[rcn]<-1
      foreach (n,v in r.costs) res[n]-=v
      sim_crafting(rc, cr, res, result)
    }
  }
}


function test_crafting()
{
  local crafted={ castle=1 fireball=1 }

  local resources={}, totalRes={}
  foreach (n in resource_names)
  {
    resources[n]<-0
    totalRes[n]<-0
  }

  print("==========================\n")

  local maxTier=0
  foreach (r in crafting_recipes)
  {
    foreach (n,v in r.costs) totalRes[n]+=v
    if (r.tier>maxTier) maxTier=r.tier
  }

  for (local tier=0; tier<=maxTier; ++tier)
  {
    print(format("tier %2d:", tier))
    foreach (n,r in crafting_recipes)
      if (r.tier==tier) print(" "+n)
    print("\n")
  }

  print("total resources needed:")
  foreach (n,v in totalRes) print(" "+n+"="+v)
  print("\n")

  local log={}
  sim_crafting(0, crafted, resources, log)
  print(log.len()+" variants:\n");

  local required=[
    ""
    "waterGlobe pit"
    ""
    ""
  ]

  foreach (ci,reqStr in required)
  {
    print("=== cluster "+ci+"\n")
    local req=split(reqStr, " ")
    foreach (n,c in log)
    {
      if (c!=ci) continue;
      print(" "+n+"\n")
      local cr={}
      foreach (cn in split(n, " ")) cr[cn]<-1

      local ok=true
      foreach (r in req) if (!(r in cr)) { ok=false; break; }

      if (!ok)
      {
        print("    Error: missing")
        foreach (r in req) if (!(r in cr)) print(" "+r)
        print("\n")
      }
    }
  }

  print("==========================\n")
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class CraftingAltarListener extends Listener
{
  </persistent=1/> id=0
  </persistent=1/> posX=0
  </persistent=1/> posZ=0

  </persistent=1/> recipeType=null
  </persistent=1/> recipeName=null
  </persistent=1/> recipeCosts=null

  </persistent=1/> craftedObj=null

  constructor(i, topo, params=null)
  {
    id=i
    posX=topo.ox
    posZ=topo.oz

    base.constructor(params)
    // craftedObj=SpellObject(posX, posZ, "FireballSpell") //==
  }

  function processEvent(e)
  {
    if (e=="crafting_on_"+id)
    {
      if (recipeType)
      {
        reset_crafting_ui()

        local canCraft=true
        foreach (n,v in recipeCosts)
        {
          add_crafting_cost(n, v)
          if (v>get_glob_res_amount(n)) canCraft=false
        }

        show_crafting_ui(get_spell_title(recipeName), get_spell_icon_name(recipeName))

        if (craftedObj) { craftedObj.destroy(); craftedObj=null }

        if (canCraft)
        {
          craftedObj=SpellObject(posX, posZ, recipeName)
          craftedObj.setEvent("crafting_get_"+id)
        }
      }
      else
        hide_crafting_ui()
    }
    else if (e=="crafting_off_"+id)
    {
      hide_crafting_ui()
      if (craftedObj) { craftedObj.destroy(); craftedObj=null }
    }
    else if (e=="crafting_get_"+id)
    {
      if (recipeType)
      {
        foreach (n,v in recipeCosts) universe.resources[n]-=v
        recipeType=null
        recipeName=null
        recipeCosts=null
        universe.craftingSeed=irnd()
      }
      craftedObj=null
      hide_crafting_ui()
    }
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class CraftingController extends Listener
{
  </persistent=1/> numAltars=0

  constructor(n, params=null)
  {
    numAltars=n
    base.constructor(params)
  }

  function processEvent(e)
  {
    if (e=="enter_world")
    {
      local recipes=get_next_crafting_recipes(numAltars)
      local i=0
      for (; i<recipes.len(); ++i)
      {
        local altar=find_listener(@(l) (l instanceof CraftingAltarListener) && l.id==i)
        if (!altar) continue;

        local r=recipes[i]
        altar.recipeType=r.type
        altar.recipeName=r.name
        altar.recipeCosts=r.costs
      }

      for (; i<numAltars; ++i)
      {
        local altar=find_listener(@(l) (l instanceof CraftingAltarListener) && l.id==i)
        if (!altar) continue;

        altar.recipeType=null
        altar.recipeName=null
        altar.recipeCosts=null
      }
    }
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class CraftingAltar extends WorldElement
{
  tags="CraftingAltar"

  id=0
  rad=10

  sensor=null
  listener=null

  constructor(i, params=null)
  {
    id=i
    base.constructor(params)

    topo=TopoCircle(rad*1.2)
    geo=GeoCircle(topo, rad*1.2, { priority=1000 })
  }

  function placeObjects()
  {
    create_stone_circle(geo.posX(), geo.posZ(), rad)
    local o=StoneCup1Object(geo.posX(), geo.posZ())
    o.makeSpecial()
    o.setActive(false)

    sensor=Sensor(vec4f(geo.posX(), 0, geo.posZ(), 1), rad)
    sensor.setOnEvent("crafting_on_"+id)
    sensor.setOffEvent("crafting_off_"+id)

    mark_circle(geo.posX(), geo.posZ(), rad, MARK_PLACE, 0)

    listener=CraftingAltarListener(id, topo)
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class CraftingWorldGen extends GenericWorldGen
{
  tags="WorldGen crafting_world normal"

  function generate()
  {
    WELEM("!PlayerStart")().addToWGen(this, "top")

    local portalPlace=TopoCircle(10)
    addElem(portalPlace, "topTopo")

    local numAltars=5
    for (local i=0; i<numAltars; ++i)
      WELEM("!CraftingAltar")(i).addToWGen(this, "altars")

    addElem(TopoGroup(getGroup("altarsTopo")), "topTopo")
    getGroup("topWE").extend(getGroup("altarsWE"))

    local shaper=WELEM("!WorldShaper")()
    shaper.shapeWorld(this)

    WorldPortal(vec4f(portalPlace.ox, 0, portalPlace.oz, 1))
    CraftingController(numAltars)
  }
}
