

class DarkWorldEruptions extends Listener
{
  </persistent=1/> mask=null
  </persistent=1/> timer=2

  function processEvent(e)
  {
    if (e=="update")
    {
      timer-=update_event_period
      if (timer<=0)
      {
        timer=rndf(5, 15)
        mask.placePoints(1, function(x, z)
        {
          add_magic_volcano(vec4f(x+srnd()*0.5, 0, z+srnd()*0.5, 1),
            rndf(2, 3), rndf(3, 10), rndf(15, 25), 1)
        })
      }
    }
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class DarkStoneListener extends Listener
{
  </persistent=1/> stoneObj=null
  </persistent=1/> sensor=null
  </persistent=1/> timer=0

  constructor(o, params=null)
  {
    base.constructor(params)
    stoneObj=o

    sensor=Sensor(stoneObj.getPos(), 30)
    sensor.use3dDistance(true)
  }

  function processEvent(e)
  {
    if (e=="update")
    {
      sensor.setPos(stoneObj.getPos())

      if (timer>0) timer-=update_event_period

      if (timer<=0 && sensor.isActive())
      {
        timer=rndf(3, 5)
        cast_lightning(stoneObj.getPos()+vec4f(0, 3, 0, 0),
          make_dir_xz(frnd())+vec4f(0, 1, 0, 0),
          sensor.getRadius(), -2,
          2, null, SE_lightning_cast)
      }
    }
  }
}


class DarkStoneFiller extends PlacedFillerElem
{
  tags="FillerElem dark -normal"

  rad=null
  avgh=0

  sensor=null

  constructor(params=null)
  {
    base.constructor(params)

    if (rad==null) rad=rndf(3, 5)

    topo=TopoCircle(rad*3.2)
    geo=GeoCircle(topo, rad*1.2, { avgh=avgh, priority=1000 })
  }

  function placeObjects()
  {
    local cpos=vec4f(geo.posX(), avgh, geo.posZ(), 1);
    local o=(frnd()<0.5?Menhir1StaticObject:Menhir2StaticObject)(getx(cpos), getz(cpos))

    DarkStoneListener(o)

    mark_circle(geo.posX(), geo.posZ(), rad, MARK_PLACE, 0)
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class LavaLakeFiller extends FloodFillerElem
{
  tags="FillerElem lava dark -normal"

  border=1

  constructor(params=null)
  {
    area=rndi(400, 800)
    base.constructor(params)
  }

  function placeFiller(fmask)
  {
    base.placeFiller(fmask)
    if (!mask) return;

    geo=GeoLake(TopoMaskHolder(mask), {priority=1100, useLava=true, depthScale=4})
    wgen.addGeo(geo)
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class DarkFinalFiller extends WorldElement
{
  tags="FinalFiller dark"

  targetMask=null

  function placeObjects()
  {
    local mask=targetMask
    if (mask==null) return;

    local area=mask.area()
    if (area==0) return;

    DarkWorldEruptions({mask=mask})

    local num=round(area/200.0)
    if (num<=0) return;

    mask.placePoints(num, function(x, z)
    {
      if (frnd()<0.1)
      {
        local o=TreeObject(x, z,
          get_pal_color(cur_world_pal, "leaves"),
          get_pal_color(cur_world_pal, "trunk"))
        o.setFoliage(false)
      }
      else
      {
        Stone1StaticObject(x, z)
      }
    })
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class DarkTrapListener extends SpawnListener
{
  function spawnObject(x, z)
  {
    cast_lightning(get_land_pos_at(vec4f(x, 0, z, 1)),
      make_dir_xz(frnd())+vec4f(0, 4, 0, 0),
      40, -2,
      30.0/number, null, SE_lightning_cast)
  }
}


class DarkTrapCircle extends WorldElement
{
  tags="MiniPuzzle Circle stone megalith dark -normal"

  rad=null
  number=null
  avgh=0
  event=null

  sensor=null

  constructor(params=null)
  {
    base.constructor(params)

    if (rad==null) rad=rndf(8, 20)

    if (number==null)
    {
      number=(rad*(0.5+srnd()*0.1)).tointeger()
      if (number<3) number=3
    }

    topo=TopoCircle(rad*1.2)
    geo=GeoCircle(topo, rad*1.2, { avgh=avgh, priority=1000 })
  }

  function placeObjects()
  {
    local objs=[]

    local cpos=vec4f(geo.posX(), avgh, geo.posZ(), 1)
    local ofs=frnd()
    for (local i=0; i<number; ++i)
    {
      local p = cpos + make_dir_xz((i+srnd()*0.1)/number+ofs) * (rad*(1+srnd()*0.04));
      local o=(frnd()<0.5?Menhir1StaticObject:Menhir2StaticObject)(getx(p), getz(p))
      o.makeSpecial()
      objs.append(o)
    }

    sensor=Sensor(cpos, rad)
    sensor.setOnEvent(event)

    DarkTrapListener(event, {topo=topo, number=rndi(15, 30)})
    MarkInactiveListener(event, objs)

    mark_circle(geo.posX(), geo.posZ(), rad, MARK_PLACE, 0)
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class InfiniteSpawnerListener extends Listener
{
  </persistent=1/> minNumber=30
  </persistent=1/> timer=0
  </persistent=1/> spawnPeriod=1
  </persistent=1/> idlePeriod=90
  </persistent=1/> mask=null

  topo=null
  maskName="mask"

  constructor(params=null)
  {
    base.constructor(params)

    if (topo!=null)
    {
      mask=topo[maskName].shiftedCopy(topo.ox, topo.oz)
      topo=null
    }
  }

  function processEvent(e)
  {
    if (e=="update")
    {
      if (timer>0) timer-=update_event_period
      else if (count_alive_monsters()<minNumber)
      {
        if (mask!=null) mask.placePoints(1, spawnObject.bindenv(this))
        timer=spawnPeriod
      }
      else timer=idlePeriod
    }
  }

  function spawnObject(x, z)
  {
    if (!monster_spawn_enabled) return;
    local p=vec4f(x, get_land_height(x, z)-5, z, 1)
    local d=make_dir_xz(frnd())
    local r=frnd()
    if (r<0.25) WyvernMonster(p-vec4f(0, 5, 0, 0))
    else if (r<0.50) LandMonster(p, d)
    else BirdMonster(p, d).setBodyModel("bird", 0xffff1100)
  }
}


class InfiniteSpawner extends WorldElement
{
  tags="InfiniteSpawner"

  area=null
  listener=null

  constructor(params=null)
  {
    base.constructor(params)

    area=WELEM("!Area")()
  }

  function addToWGen(wg, prefix)
  {
    area.addToWGen(wg, prefix)
    wg.addElem(this, prefix+"WE")
  }

  function placeObjects()
  {
    listener=InfiniteSpawnerListener({ topo=area.topo })
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class DarkWorldGen extends GenericWorldGen
{
  tags="WorldGen generic dark"

  function generate()
  {
    set_sky_name("dark"+rndi(1, 3))

    local amap=HmapMask.fromSimplexNoise(1, 8, 6, null, 0.5)
    local ak=rndf(0.8, 1)
    ashMap=HmapMask(@(x, z) (amap.getraw(x, z)+1)*ak)

    for (local n=rndi(1, 3); n>0; --n)
      WELEM("!InfiniteSpawner")().addToWGen(this, "top")

    base.generate()

    ShowMessageListener("enter_world",
      {text="This looks dangerous. I might need to escape\nby using the portal spell on my castle."})

    hubTrigger.maxMonsters=10
  }
}
