

class GenericFiller extends WorldElement
{
  tags="Filler"

  targetTopo=null
  targetMaskName="mask"
  areaThreshold=0.75

  elems=null

  function afterTopo()
  {
    elems=[]

    local mask=null
    if (targetTopo) mask=targetTopo[targetMaskName].shiftedCopy(targetTopo.ox, targetTopo.oz)

    if (mask==null || mask.area()==0) return;

    local thr=mask.area()*areaThreshold
    while (mask.area()>thr)
    {
      local e=WELEM("!FillerElem")()
      e.placeFiller(mask)
      elems.append(e)
    }

    elems.append(WELEM("!FinalFiller")({targetMask=mask}))
  }

  function placeObjects()
  {
    foreach (e in elems) e.placeObjects()
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class PlacedFillerElem extends WorldElement
{
  border=0

  function placeFiller(fmask)
  {
    topo.makeMask()

    local self=this
    fmask.placeCircles([topo.mask.boundingCircle()[2]], function(x, z)
    {
      local bc=self.topo.mask.boundingCircle()
      self.topo.setPos(x-bc[0], z-bc[1])
    })

    if (geo) wgen.addGeo(geo)

    local m=topo.mask
    for (local iter=border; iter>0; --iter) m=m.expand()

    fmask.bitwiseInplace(m, 2, topo.ox, topo.oz)
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class FloodFillerElem extends WorldElement
{
  border=0
  area=null

  mask=null

  constructor(params=null)
  {
    base.constructor(params)
    if (area==null) area=rndi(200, 800)
  }

  function placeFiller(fmask)
  {
    local i=fmask.randomPixel()
    if (i<0) return;

    mask=fmask.randomFill(map_index_x(i), map_index_z(i), area)

    for (local iter=border; iter>0; --iter)
    {
      mask=mask.expand()
      mask.bitwiseInplace(fmask, 8, 0, 0)
    }

    fmask.bitwiseInplace(mask, 2, 0, 0)
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class StoneFiller extends FloodFillerElem
{
  tags="FillerElem stone normal dark"

  function placeObjects()
  {
    if (!mask) return;

    for (local iter=3; iter>0; --iter) mask=mask.erode()

    local num=round(mask.area()/30.0)
    if (num<=0) return;

    mask.placePoints(num, function(x, z)
    {
      Stone1StaticObject(x+srnd()*0.5, z+srnd()*0.5)
    })
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class ForestFiller extends FloodFillerElem
{
  tags="FillerElem tree normal -dark"

  constructor(params=null)
  {
    area=rndi(700, 2000)
    base.constructor(params)
  }

  function placeObjects()
  {
    if (!mask) return;

    for (local iter=3; iter>0; --iter) mask=mask.erode()

    local num=round(mask.area()/100.0)
    if (num<=0) return;

    mask.placePoints(num, function(x, z)
    {
      local o=TreeObject(x, z,
        get_pal_color(cur_world_pal, "leaves"),
        get_pal_color(cur_world_pal, "trunk"))
      o.setFoliage(true)
    })
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class LakeFiller extends FloodFillerElem
{
  tags="FillerElem water normal -islands"
  weight=20

  border=1

  constructor(params=null)
  {
    area=rndi(300, 800)
    base.constructor(params)
  }

  function placeFiller(fmask)
  {
    base.placeFiller(fmask)
    if (!mask) return;

    geo=GeoLake(TopoMaskHolder(mask), {priority=1100, useLava=false, depthScale=1})
    wgen.addGeo(geo)
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class GenericFinalFiller extends WorldElement
{
  tags="FinalFiller normal"

  targetMask=null

  function placeObjects()
  {
    local mask=targetMask
    if (mask==null) return;

    local area=mask.area()
    local num=round(area/200.0)
    if (num<=0) return;

    // print(format("!!! area %d, about to place %d objects\n", area, num))

    mask.placePoints(num, function(x, z)
    {
      if (frnd()<0.5)
      {
        local o=TreeObject(x, z,
          get_pal_color(cur_world_pal, "leaves"),
          get_pal_color(cur_world_pal, "trunk"))
      }
      else
      {
        Stone1StaticObject(x, z)
      }
    })
  }
}
