

universe_devmode <- false


resource_names <- ["shard" "vortex" "firebloom" "star" "seed"]


function make_world(x, y, color)
{
  local scheme="generic normal"

  local mr={}, br={}

  if (frnd()<0.1) mr.portal<-1
  else br.portal<-1

  return {
    ux=x
    uy=y
    seed=irnd()
    scheme=scheme
    mainRewards=mr
    bonusRewards=br
    remainingRewards={}
    hidden=true
    recent=false
    activated=false
    visited=false
    color1=color
    // color2=color
    color2=tocolor(hsv2rgba(frnd()*360, rndf(0.5, 1), rndf(0.4, 1), 1))
    description=null
    g={
      resRewards=true
    }
  }
}


function world_has_tag(s, tag)
{
  foreach (t in split(s, " \t")) if (t==tag) return true;
  return false;
}


function make_link(a, b)
{
  if (a<0 || b<0) return;
  local l={ from=a, to=b, type="bi", internal=true }
  universe.links.append(l)
  return l;
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


class UniCluster
{
  topo=null
  maskName="mask"
  numWorlds=0
  worlds=null
  color=0
  resources=null
  tags=null


  constructor(h, n, params=null)
  {
    numWorlds=n
    color=tocolor(hsv2rgba(h*360, rndf(0.5, 1), rndf(0.4, 0.7), 1))
    worlds=[]
    resources={}

    if (params) foreach (n,v in params)
      if (n in this) this[n]=v

    if (!topo) topo=TopoBlot(numWorlds*2, {uniformSteps=0})
  }


  function createWorlds(ox, oz)
  {
    local gridRnd=0.7
    local gridStep=30

    ox+=topo.ox
    oz+=topo.oz

    topo[maskName].placePoints(numWorlds, function(x, z)
    {
      local w=make_world(
        (wrap(x+ox)+frnd()*gridRnd)*gridStep,
        (wrap(z+oz)+frnd()*gridRnd)*gridStep,
        color)
      if (tags) w.scheme=tags
      worlds.append(w)
    }.bindenv(this))
  }


  function findNearestWorlds(other)
  {
    local bd=1e20, bw1=null, bw2=null

    foreach (w in worlds)
      foreach (ow in other.worlds)
      {
        local dx=w.ux-ow.ux
        local dy=w.uy-ow.uy
        local d=dx*dx+dy*dy
        if (d>=bd) continue;
        bd=d
        bw1=w
        bw2=ow
      }

    return [bw1, bw2];
  }


  function findNearestWorld(ow)
  {
    local bd=1e20, bw=null

    foreach (w in worlds)
    {
      local dx=w.ux-ow.ux
      local dy=w.uy-ow.uy
      local d=dx*dx+dy*dy
      if (d>=bd) continue;
      bd=d
      bw=w
    }

    return bw;
  }


  function linkCluster(c)
  {
    local w=findNearestWorlds(c)
    make_link(w[0].id, w[1].id).internal=false
    return w;
  }


  function linkWorld(ow)
  {
    local w=findNearestWorld(ow)
    make_link(w.id, ow.id).internal=false
    return w;
  }


  function getGenericWorlds(num)
  {
    local filtered=[]
    foreach (w in worlds) if (world_has_tag(w.scheme, "generic")) filtered.append(w)

    local res=[]
    for (; num>0 && filtered.len(); --num)
    {
      local i=rndi(0, filtered.len()-1)
      res.append(filtered[i])
      filtered.remove(i)
    }

    return res;
  }


  function createInternalLinks()
  {
    for (local i=0; i<worlds.len(); ++i)
    {
      local wi=worlds[i]
      for (local j=i+1; j<worlds.len(); ++j)
      {
        local wj=worlds[j]
        local dx=wi.ux-wj.ux
        local dy=wi.uy-wj.uy
        local dij=dx*dx+dy*dy

        local ok=true
        foreach (k,w in worlds)
        {
          if (k==i || k==j) continue;

          dx=w.ux-wi.ux
          dy=w.uy-wi.uy
          local d=dx*dx+dy*dy
          if (d>=dij) continue;

          dx=w.ux-wj.ux
          dy=w.uy-wj.uy
          d=dx*dx+dy*dy
          if (d>=dij) continue;

          ok=false
          break;
        }

        if (ok) make_link(wi.id, wj.id)
      }
    }
  }


  function pickResource()
  {
    local sum=0
    foreach (v in resources) sum+=v
    if (sum<=0) return null;

    local r=rndi(0, sum-1)
    local a=0
    foreach (n,v in resources)
    {
      if (v<=0) continue;
      a+=v
      if (r<a) return n;
    }

    return null;
  }


  function placeResources()
  {
    local filtered=[]
    foreach (w in worlds) if (w.g.resRewards) filtered.append(w)

    // place one main reward in each world
    foreach (w in filtered)
    {
      local r=pickResource()
      if (!r) break;

      if (r in w.mainRewards) w.mainRewards[r]++
      else w.mainRewards[r]<-1

      resources[r]--
    }

    // place remaining resource rewards
    for (;;)
    {
      local r=pickResource()
      if (!r) break;

      local w=filtered[rndi(0, filtered.len()-1)]
      local rw=w.mainRewards
      if (frnd()<0.5) rw=w.bonusRewards

      if (r in rw) rw[r]++
      else rw[r]<-1

      resources[r]--
    }
  }
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


function generate_universe()
{
  // create clusters
  local hue=frnd()

  local cluster1=UniCluster(hue, rndi(10, 14),
    {resources={ shard=20 vortex=20 }, tags="generic normal -tree_puzzle"})
  hue=(hue+GOLD_RATIO)%1

  local dark1=UniCluster(hue, 2,
    {resources={ vortex=5 seed=1 }, tags="generic dark",
    color=tocolor(hsv2rgba(hue*360, rndf(0.5, 1), 0.1, 1)),
    })
  hue=(hue+GOLD_RATIO)%1

  local gateway1=UniCluster(hue, 1)

  local clusters=[cluster1, dark1, gateway1]
  local grp=[TopoGroup([cluster1.topo, gateway1.topo, dark1.topo])]
  local cluster2=null
  local outer=[]

  if (DEMO)
  {
    local c=UniCluster(hue, 1)
    clusters.append(c)
    // outer.append(c)
    grp.append(c.topo)
    cluster2=c
  }
  else
  {
    for (local n=3; n>0; --n)
    {
      local c=UniCluster(hue, rndi(10, 15),
        {resources={ shard=30 vortex=12 firebloom=20 seed=12 star=8 }})
      hue=(hue+GOLD_RATIO)%1
      clusters.append(c)
      outer.append(c)
      grp.append(c.topo)
      if (!cluster2) cluster2=c
    }
  }

  local topo=TopoIsland(TopoGroup(grp), {extraSteps=0, uniformSteps=2})

  local dark2=UniCluster(hue, rndi(15, 25),
    {resources={ shard=60 vortex=24 firebloom=40 seed=24 star=16 },
    tags="generic dark",
    color=tocolor(hsv2rgba(hue*360, rndf(0.5, 1), 0.1, 1)),
    topo=topo, maskName="maskAround"
    })
  clusters.append(dark2)

  topo.makeMask()

  // create worlds
  local center=topo.getCenter()

  local worlds=[]
  foreach (c in clusters)
  {
    c.createWorlds(-getx(center), -getz(center))
    worlds.extend(c.worlds)
  }

  foreach (i,w in worlds) w.id<-i

  universe=
  {
    worlds=worlds,
    links=[],
    focusWorld=-1,
    enteredWorld=-1,
    craftingSeed=irnd(),
  }

  for (local n=5; n; --n) relax_worlds_pos()

  // link clusters
  local w=gateway1.linkCluster(cluster1)
  gateway1.linkCluster(cluster2)
  w[0].description="A strange world. It looks like there's\nsome kind of puzzle here."
  w[0].scheme="tree_puzzle_world normal"
  w[0].g.resRewards=false
  w[0].mainRewards={}
  w[0].bonusRewards={}
  w[0].color1=w[0].color2=0xffffffff

  foreach (w in dark1.worlds)
  {
    w.description="A dark world."
    cluster1.linkWorld(w)
  }

  foreach (c in outer)
  {
    dark2.linkCluster(c)
    if (c!=cluster2) cluster2.linkCluster(c)
  }

  foreach (w in dark2.worlds)
    w.description="A dark world."

  // pick special worlds
  local w=cluster1.getGenericWorlds(2)

  w[0].description="This is where I started."
  w[0].scheme="starting_world normal"
  w[0].hidden=false
  universe.focusWorld=w[0].id
  w[0].g.resRewards=false
  w[0].mainRewards={}
  w[0].bonusRewards={}

  w[1].description="I can make new spells from ingredients here."
  w[1].scheme="crafting_world normal"
  w[1].g.resRewards=false
  w[1].mainRewards={}
  w[1].bonusRewards={}
  w[1].color1=w[1].color2=0xff44ff11

  if (DEMO)
  {
    local w=cluster2.worlds[0]
    w.description="That's the end of this tiny DEMO universe!"
    w.scheme="end_of_demo normal"
    w.g.resRewards=false
    w.mainRewards={}
    w.bonusRewards={}
    w.color1=w.color2=0xffff6611
  }

  // finish generation
  foreach (c in clusters) c.createInternalLinks()
  foreach (c in clusters) c.placeResources()

  // clean up
  foreach (w in worlds)
  {
    w.remainingRewards<-{}

    foreach (n,v in w.mainRewards)
      if (!(n in w.remainingRewards)) w.remainingRewards[n]<-v
      else w.remainingRewards[n]+=v

    foreach (n,v in w.bonusRewards)
      if (!(n in w.remainingRewards)) w.remainingRewards[n]<-v
      else w.remainingRewards[n]+=v

    delete w.g
  }

  print(format("%d worlds total\n", worlds.len()))
  return universe;
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


function get_world_tooltip(w)
{
  local s=""

  if (universe_devmode)
  {
    s+="World #"+w.id+" seed "+w.seed+"\n"
    s+=w.scheme
    s+="\n---"

    // s+="\nmain rewards:"
    // foreach (n,v in w.mainRewards) s+=" "+n+"="+v
    // s+="\nbonus rewards:"
    // foreach (n,v in w.bonusRewards) s+=" "+n+"="+v
    s+="\nremaining rewards:"
    foreach (n,v in w.remainingRewards) s+=" "+n+"="+v
    s+="\n---\n"
  }

  if (!w.visited) s+="Unvisited.\n"
  else if (w.description) s+=w.description+"\n"
  if (w.recent) s+="Recently discovered.\n"
  if (w.activated) s+="Portal is active.\n"
  if (w.visited && w.remainingRewards.len()>0) s+="There are unseen treasures here.\n"

  return s;
}


//ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ//


function relax_worlds_pos()
{
  local old=array(universe.worlds.len())
  for (local i=0; i<old.len(); ++i) old[i]=clone universe.worlds[i]

  local rk=0.1

  foreach (w in universe.worlds)
  {
    local cx=w.ux, cy=w.uy
    local bx=0, by=0, bd=60
    bd*=bd
    foreach (ow in old)
    {
      if (ow.id==w.id) continue;
      local dx=ow.ux-cx, dy=ow.uy-cy
      local d=dx*dx+dy*dy
      if (d>bd) continue;
      bd=d; bx=dx; by=dy
    }

    w.ux-=bx*rk
    w.uy-=by*rk
  }
}


function unhide_world(w)
{
  if (!w.hidden) return;
  w.hidden=false
  w.recent=true
}


function activate_world(w)
{
  if (w.id<0) return;

  universe.focusWorld=w.id
  w.hidden=false
  w.recent=false

  if (w.activated) return;
  w.activated=true

  foreach (w in universe.worlds) w.recent=false

  foreach (l in universe.links)
  {
    if (l.type=="bi")
    {
      if (l.from==w.id || l.to==w.id)
      {
        unhide_world(universe.worlds[l.from])
        unhide_world(universe.worlds[l.to  ])
      }
    }
  }
}


function activate_current_world()
{
  if (!cur_world || cur_world.id<0) return;

  activate_world(cur_world)
  // universe.enteredWorld=-1
  save_game()
}


function get_world_ctx_menu(w)
{
  if (!universe_devmode) return null;

  local menu=[]
  menu.append("World #"+w.id)
  if (w.hidden) menu.append(["Unhide", function() { w.hidden=false }])
  menu.append(["Toggle visited", function() { w.visited=!w.visited }])
  menu.append(["Activate", function() { activate_world(w); save_game() }])
  // menu.append(["relax", relax_worlds_pos])
  return menu;
}
